"""batou templating support

Currently we support one templating engine:

Jinja2::
    {% for server in servers %}
    server {{ server.name }}
    {% endfor %}

"""

from __future__ import print_function, unicode_literals
import jinja2
import StringIO
from batou import output


class TemplateEngine(object):
    """Abstract templating wrapper class.

    Use a subclass that connects to a specific template engine.
    """

    @classmethod
    def get(cls, enginename):
        """Return TemplateEngine instance for `enginename`."""
        if enginename.lower() == 'jinja2':
            return Jinja2Engine()
        raise NotImplementedError('template engine not known', enginename)

    def template(self, sourcefile, args):
        """Render template from `sourcefile` and return the value.
        """
        return self._render_template_file(sourcefile, args).getvalue()

    def _render_template_file(self, sourcefile, args):
        """Expand template found in `sourcefile` and return it as StringIO."""
        raise NotImplementedError

    def expand(self, templatestr, args):
        """Expand template in `templatestr` and return it as string."""
        raise NotImplementedError


class Jinja2Engine(TemplateEngine):

    def __init__(self, *args, **kwargs):
        super(Jinja2Engine, self).__init__(*args, **kwargs)
        self.env = jinja2.Environment(
            line_statement_prefix='@@',
            keep_trailing_newline=True,
            undefined=jinja2.StrictUndefined)

    def _render_template_file(self, sourcefile, args):
        tmpl = open(sourcefile).read()
        tmpl = self.env.from_string(tmpl)
        output = StringIO.StringIO()
        print(tmpl.render(args), file=output)
        return output

    def expand(self, templatestr, args, identifier='<template>'):
        if len(templatestr) > 100*1024:
            output.error(
                "You are trying to render a template that is bigger than "
                "100KiB we've seen that Jinja can crash at large templates "
                "and suggest you find alternatives for this. The affected "
                "template starts with:")
            output.annotate(templatestr[:100])
        tmpl = self.env.from_string(templatestr)
        tmpl.filename = identifier
        return tmpl.render(**args)



from batou import UpdateNeeded, output
from batou.component import Component
from batou.lib.file import Directory, ensure_path_nonexistent
import os.path


class Clone(Component):

    namevar = 'url'
    target = '.'
    revision = None
    branch = None
    vcs_update = True

    def configure(self):
        if (not self.revision_or_branch) or (self.revision and self.branch):
            raise ValueError(
                'Clone(%s) needs exactly one of revision or branch' % self.url)
        self.target = self.map(self.target)
        self += Directory(self.target)

    def verify(self):
        self._force_clone = False
        if not os.path.exists(self.target):
            self._force_clone = True
            raise UpdateNeeded()
        # if the path does exist but isn't a directory, just let the error
        # bubble for now, the message will at least tell something useful
        with self.chdir(self.target):
            if not os.path.exists('.git'):
                self._force_clone = True
                raise UpdateNeeded()

            if self.remote_url() != self.url:
                self._force_clone = True
                raise UpdateNeeded()

            if not self.vcs_update:
                return

            if self.has_outgoing_changesets():
                output.annotate(
                    'Git clone at {} has outgoing changesets.'.format(
                        self.target))

            if self.has_changes():
                output.annotate(
                    'Git clone at {} is dirty, going to lose changes.'.format(
                        self.target), red=True)
                raise UpdateNeeded()

            if self.revision and self.current_revision() != self.revision:
                raise UpdateNeeded()
            if (self.branch and (
                    self.current_branch() != self.branch or
                    self.has_incoming_changesets())):
                raise UpdateNeeded()

    @property
    def revision_or_branch(self):
        return self.revision or self.branch

    def current_revision(self):
        try:
            with self.chdir(self.target):
                stdout, stderr = self.cmd('LANG=C git show -s --format=%H')
        except RuntimeError:
            return None
        return stdout.strip()

    def current_branch(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('git rev-parse --abbrev-ref HEAD')
        return stdout.strip()

    def has_incoming_changesets(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('git fetch --dry-run')
            return stderr.strip()

    def has_outgoing_changesets(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('LANG=C git status')
        return 'Your branch is ahead of' in stdout

    def has_changes(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('git status --porcelain')
        return bool(stdout.strip())

    def remote_url(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('git remote get-url origin')
        return stdout.strip()

    def update(self):
        just_cloned = False
        if self._force_clone:
            ensure_path_nonexistent(self.target)
            self.cmd(self.expand(
                'git clone {{component.url}} {{component.target}}'))
            just_cloned = True
        with self.chdir(self.target):
            for filepath in self.untracked_files():
                os.unlink(os.path.join(self.target, filepath))
            if not just_cloned:
                self.cmd('git fetch')
            if self.branch:
                self.cmd(self.expand(
                    'git reset --hard origin/{{component.branch}}'))
            else:
                self.cmd(self.expand(
                    'git reset --hard {{component.revision}}'))

            # XXX We should re-think submodule support; e.g. which revision
            # shall the submodules be updated to?
            self.cmd('git submodule update --init --recursive')

    def untracked_files(self):
        stdout, stderr = self.cmd(
            'git status --porcelain --untracked-files=all')
        items = (line.split(None, 1) for line in stdout.splitlines())
        return [filepath for status, filepath in items if status == '??']

    def last_updated(self):
        with self.chdir(self.target):
            if not os.path.exists('.git'):
                return None
            stdout, stderr = self.cmd('git show -s --format=%ct')
            timestamp = stdout.strip()
            return float(timestamp)



import batou.lib.supervisor
import os.path
import pytest


@pytest.yield_fixture
def supervisor(root, request):
    # Urks. Otherwise OS X ends up with socket paths that are too long.
    supervisor = batou.lib.supervisor.Supervisor(
        pidfile='supervisor.pid',
        socketpath='/tmp/batou-test-supervisor.sock')
    root.component += supervisor
    root.component.deploy()
    yield supervisor
    supervisor.cmd(
        '{}/bin/supervisorctl shutdown'.format(supervisor.workdir))


@pytest.mark.slow
def test_waits_for_start(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        'foo', command_absolute=False,
        command='bash', args='-c "sleep 1; touch %s/foo; sleep 3600"' % (
            root.workdir), options=dict(startsecs=2))
    root.component.deploy()
    assert os.path.exists('%s/foo' % root.workdir)


@pytest.mark.slow
def test_program_does_not_start_within_startsecs_raises(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        'foo', command_absolute=False, command='true',
        options=dict(startsecs=1))
    with pytest.raises(RuntimeError):
        root.component.deploy()



import sys
from batou.remote_core import Output


class TerminalBackend(object):

    def __init__(self):
        import py.io
        self._tw = py.io.TerminalWriter(sys.stdout)

    def line(self, message, **format):
        self._tw.line(message, **format)

    def sep(self, sep, title, **format):
        self._tw.sep(sep, title, **format)

    def write(self, content, **format):
        self._tw.write(content, **format)


class NullBackend(object):

    def line(self, message, **format):
        pass

    def sep(self, sep, title, **format):
        pass

    def write(self, content, **format):
        pass


output = Output(NullBackend())



from batou.secrets.edit import Editor
import mock


def test_editor_exits_on_successful_command():
    editor = Editor('vim', mock.Mock())
    editor.edit = mock.Mock()
    editor.encrypt = mock.Mock()
    editor.main()



from batou.component import Component, ComponentDefinition
from batou.environment import Environment
import batou.utils
import os
import pytest


@pytest.fixture
def root(tmpdir):
    environment = Environment('test', basedir=str(tmpdir))
    environment._set_defaults()
    os.chdir(str(tmpdir))

    class MyComponent(Component):
        pass

    compdef = ComponentDefinition(MyComponent)
    compdef.defdir = str(tmpdir)
    environment.components[compdef.name] = compdef
    root = environment.add_root(compdef.name, 'host')
    root.prepare()
    root.component.deploy()
    return root


@pytest.yield_fixture(autouse=True)
def ensure_workingdir(request):
    working_dir = os.getcwd()
    yield
    os.chdir(working_dir)


def pytest_assertrepr_compare(op, left, right):
    if left.__class__.__name__ == 'Ellipsis':
        return left.compare(right).diff
    elif right.__class__.__name__ == 'Ellipsis':
        return right.compare(left).diff


@pytest.fixture(autouse=True)
def reset_resolve_overrides():
    batou.utils.resolve_override.clear()
    batou.utils.resolve_v6_override.clear()



from batou.component import Component
from batou.lib.archive import Extract
from batou.lib.download import Download
from batou.utils import CmdExecutionError
import batou


class VirtualEnv(Component):
    """Manage a virtualenv installation.

    This also manages a clean virtualenv base installation to ensure we know
    what we're dealing with.

    The virtualenv provides a somewhat abstract interface to specify packages
    to be installed. It tries to use reasonably safe defaults for calling the
    virtualenv, pip, etc. commands.

    It also upgrades the whole virtualenv if the virtualenv version itself is
    upgraded.

    """

    # The Python major version that is requested.
    namevar = 'version'

    # A specific executable to use as the base Python installation to create a
    # virtualenv from. Default: discover they 'python{version}' binary from the
    # search path.
    executable = None

    @property
    def python(self):
        """Path to the generated python executable."""
        return 'bin/python{}'.format(self.version)

    def configure(self):
        class_name = 'VirtualEnvPy{}'.format(self.version.replace('.', '_'))
        self.venv = globals().get(class_name, VirtualEnvPy)()
        self += self.venv

        if not self.executable:
            self.executable = 'python{}'.format(self.version)


class VirtualEnvPyBase(Component):

    venv_version = None
    venv_checksum = None
    venv_options = ()

    installer = 'pip'
    install_options = ('--ignore-installed',)

    def verify(self):
        self.assert_cmd(
            'bin/python -c "import sys; '
            'assert sys.version_info[:2] == {}"'.format(repr(
                tuple(int(x) for x in self.parent.version.split('.')))))
        # Is this Python (still) functional 'enough'
        # from a setuptools/distribute perspective?
        self.assert_cmd('bin/python -c "import pkg_resources"')

    def update(self):
        self.cmd('chmod -R u+w bin/ lib/ include/ .Python || true')
        self.cmd('rm -rf bin/ lib/ include/ .Python')

    def verify_pkg(self, pkg):
        try:
            self.cmd(
                'bin/python -c "'
                'import pkg_resources; '
                'assert pkg_resources.require(\'{}\')[0].parsed_version == '
                'pkg_resources.parse_version(\'{}\')"'
                .format(pkg.package, pkg.version))
        except CmdExecutionError:
            raise batou.UpdateNeeded()
        # Is the package usable? Is the package a module?  This might be
        # overspecific - I'm looking for a way to deal with:
        # https://github.com/pypa/pip/issues/3 if a namespace package was not
        # installed cleanly. This only works (currently), when the package name
        # corresponds with what it contains. I.E. it works for zc.buildout but
        # not for distribute, which installs a setuptools package.
        if pkg.check_package_is_module:
            try:
                self.cmd('bin/python -c "import pkg_resources; '
                         'import {0};{0}.__file__"'.format(pkg.package))
            except CmdExecutionError:
                raise batou.UpdateNeeded()

    def update_pkg(self, pkg):
        if self.installer == 'pip':
            self.pip_install(pkg)
        else:
            self.easy_install(pkg)

    def pip_install(self, pkg):
        options = self.install_options
        options += pkg.install_options
        if not pkg.dependencies:
            options += ('--no-deps',)
        options = ' '.join(options)
        self.cmd('bin/pip --timeout={} install {} '
                 '"{}=={}"'.format(
                     pkg.timeout, options, pkg.package, pkg.version),
                 env=pkg.env if pkg.env else {})

    def easy_install(self, pkg):
        # this old installation of pip doesn't support eggs and completely
        # screws up namespace packages. I'm afraid we have to go for
        # easy_install in this case.
        options = self.install_options
        options += pkg.install_options
        if not pkg.dependencies:
            options += ('--no-deps',)
        options = ' '.join(options)
        # XXX does not implement timeout. could we just do this on
        # 'cmd' instead?
        self.cmd('bin/easy_install {} '
                 '"{}=={}"'.format(options, pkg.package, pkg.version),
                 env=pkg.env if pkg.env else {})


class VirtualEnvPy2_7(VirtualEnvPyBase):

    venv_version = '16.1.0'
    venv_checksum = 'md5:ad6d2ebd6885b3a2b4ff2030ce532a2f'
    venv_options = ()

    install_options = ()

    def configure(self):
        self.base = VirtualEnvDownload(
            self.venv_version,
            checksum=self.venv_checksum)
        self += self.base

    def verify(self):
        # Did we install an updated virtualenv package in between?
        self.assert_file_is_current('bin/python', [self.base.venv_cmd])
        super(VirtualEnvPy2_7, self).verify()

    def update(self):
        super(VirtualEnvPy2_7, self).update()
        self.cmd('{} {} {} --python={} {}'.format(
            self.parent.executable,
            self.base.venv_cmd,
            ' '.join(self.venv_options),
            self.parent.executable, self.workdir))


class VirtualEnvPy(VirtualEnvPyBase):
    """VirtualEnv for Python with `-m venv` (>=3.3)"""

    def update(self):
        super(VirtualEnvPy, self).update()
        self.cmd('{} -m venv {}'.format(
            self.parent.executable, self.workdir))


class VirtualEnvDownload(Component):
    """Manage virtualenv package download and extraction.

    Keeps knowledge about specific virtualenv version.
    """

    namevar = 'version'
    checksum = None
    download_url = (
        'https://github.com/pypa/virtualenv/tarball/{{component.version}}')

    def configure(self):
        # This will manage central, version-specific virtualenv base
        # installations for multiple components to share.
        self.workdir = self.environment.workdir_base + '/.virtualenv'
        self += Download(
            self.expand(self.download_url),
            target=self.version + '.tar.gz',  # so Extract knows what to do
            checksum=self.checksum)
        download = self._
        extracted_dir = 'virtualenv-' + self.version
        self += Extract(download.target, target=extracted_dir, strip=1)
        self.venv_cmd = (
            self.workdir + '/' + extracted_dir + '/src/virtualenv.py')

    def verify(self):
        self.assert_no_subcomponent_changes()

    def update(self):
        self.touch(self.venv_cmd)


class Package(Component):
    """Install a package into a virtual python environment.

    Assumes the parent component is a virtual env component.

    """

    namevar = 'package'
    version = None
    check_package_is_module = True
    timeout = None
    dependencies = True
    env = None

    # Additional options to pass to the installer. Installer depends on venv.
    install_options = ()

    def configure(self):
        if not isinstance(self.parent, VirtualEnv):
            raise TypeError(
                'Package() must be added to a virtual environment')
        if self.timeout is None:
            self.timeout = self.environment.timeout

    # Actual verify/update is delegated the specific virtualenv implementation.
    def verify(self):
        self.parent.venv.verify_pkg(self)

    def update(self):
        self.parent.venv.update_pkg(self)

    @property
    def namevar_for_breadcrumb(self):
        return '{}=={}'.format(self.package, self.version)



from batou.component import Attribute, Component, platform
from batou.lib.cron import CronJob
import batou.lib.logrotate
import batou.lib.service
import batou.lib.supervisor


@platform('debian', batou.lib.service.Service)
class RebootCronjob(Component):

    def configure(self):
        self += CronJob(
            self.expand(
                '{{component.root.workdir}}/{{component.parent.executable}}'),
            timing='@reboot', logger=self.root.name)


# XXX can't use @platform since that's too late (see #12418)
class Supervisor(batou.lib.supervisor.Supervisor):

    pidfile = Attribute(str, 'var/supervisord.pid', map=True)


class Logrotate(batou.lib.logrotate.Logrotate):

    common_config = """\
daily
rotate 14
create
dateext
compress
notifempty
nomail
noolddir
missingok
sharedscripts
"""


@platform('debian', Logrotate)
class LogrotateCronjob(Component):

    def configure(self):
        self.directory = self.parent.workdir
        self += CronJob(
            self.expand(
                '/usr/sbin/logrotate -s {{component.directory}}/state'
                ' {{component.directory}}/logrotate.conf'),
            timing='45 2 * * *')



import os
import pytest
import shutil


@pytest.fixture()
def sample_service(tmpdir):
    shutil.copytree(os.path.dirname(__file__)+'/fixture/sample_service',
                    str(tmpdir / 'sample_service'))
    target = str(tmpdir / 'sample_service')
    os.chdir(target)
    return target



from batou.utils import cmd
import batou.lib.git
import os.path
import pytest


@pytest.fixture(scope='function')
def repos_path(root, name='upstream'):
    repos_path = os.path.join(root.environment.workdir_base, name)
    cmd('mkdir {dir}; cd {dir}; git init;'
        'git config user.name Jenkins;'
        'git config user.email jenkins@example.com;'
        'touch foo; git add .;'
        'git commit -am "foo"'.format(
            dir=repos_path))
    return repos_path


@pytest.mark.slow
def test_runs_git_to_clone_repository(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/foo'))
    root.component.deploy()  # trigger verify


@pytest.mark.slow
def test_directly_after_clone_nothing_is_merged(root, repos_path):
    # When a Clone is confirgured with a branch, the general gesture for
    # updating is "git merge origin/branch", but directly after cloning, the
    # Clone is still on master, so this would try to merge the configured
    # branch into master, which is wrong.
    cmd('cd {dir}; git checkout -b other; touch bar; echo qux > foo;'
        'git add .; git commit -am "other";'
        # Set up branches to be different, so we see that no merge takes place
        'git checkout master; '
        'echo one > foo; git add . ; git commit -am "foo master";'.format(
            dir=repos_path))
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='other')
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/bar'))


@pytest.mark.slow
def test_setting_branch_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    cmd('cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
        dir=repos_path))
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/bar'))


@pytest.mark.slow
def test_setting_revision_updates_on_incoming_changes(root, repos_path):
    cmd('cd {dir}; touch bar; git add .; git commit -m "commit2"'.format(
        dir=repos_path))
    commit1, _ = cmd('cd {dir}; git rev-parse HEAD^'.format(dir=repos_path))
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', revision=commit1)
    root.component.deploy()
    cmd('cd {dir}; touch qux; git add .; git commit -m "commit3"'.format(
        dir=repos_path))
    root.component.deploy()  # Our main assertion: Nothing breaks here
    assert not os.path.isfile(
        os.path.join(root.environment.workdir_base, 'mycomponent/clone/qux'))


@pytest.mark.slow
def test_branch_does_switch_branch(root, repos_path):
    cmd('cd {dir}; touch bar; git add .; git checkout -b bar;'
        'git commit -m "commit branch"'.format(dir=repos_path))
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='bar')
    root.component.deploy()
    stdout, stderr = cmd(
        'cd {workdir}/clone; git rev-parse --abbrev-ref HEAD'.format(
            workdir=root.workdir))
    assert 'bar' == stdout.strip()


@pytest.mark.slow
def test_has_changes_counts_changes_to_tracked_files(root, repos_path):
    clone = batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd('touch {}/clone/bar'.format(root.workdir))
    cmd('cd {}/clone; git add bar'.format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_has_changes_counts_untracked_files_as_changes(root, repos_path):
    clone = batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd('touch {}/clone/bar'.format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_clean_clone_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    cmd('cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
        dir=repos_path))
    root.component.deploy()
    assert os.path.isfile(root.component.map('clone/bar'))


@pytest.mark.slow
def test_changes_lost_on_update_with_incoming(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    cmd('cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
        dir=repos_path))
    cmd('cd {dir}/clone; echo foobar >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert os.path.exists(root.component.map('clone/bar'))
    assert not open(root.component.map('clone/foo')).read()


@pytest.mark.slow
def test_changes_lost_on_update_without_incoming(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    cmd('cd {dir}/clone; echo foobar >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert not open(root.component.map('clone/foo')).read()


@pytest.mark.slow
def test_untracked_files_are_removed_on_update(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master')
    root.component.deploy()
    cmd('cd {dir}/clone; mkdir bar; echo foobar >bar/baz'.format(
        dir=root.workdir))
    root.component.deploy()
    assert not os.path.exists(root.component.map('clone/bar/baz'))


@pytest.mark.slow
def test_clean_clone_vcs_update_false_leaves_changes_intact(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target='clone', branch='master', vcs_update=False)
    root.component.deploy()
    cmd('cd {dir}; echo foobar >foo; touch bar; git add .; '
        'git commit -m "commit"'.format(dir=repos_path))
    cmd('cd {dir}/clone; echo asdf >foo'.format(dir=root.workdir))
    root.component.deploy()
    assert 'asdf\n' == open(root.component.map('clone/foo')).read()
    assert not os.path.exists(root.component.map('clone/bar'))


@pytest.mark.slow
def test_changed_remote_is_updated(root, repos_path):
    git = batou.lib.git.Clone(repos_path, target='clone', branch='master')
    root.component += git

    # Fresh, unrelated repos
    repos2_path = globals()['repos_path'](root, name='upstream2')
    cmd('cd {dir}; echo baz >bar; git add .;'
        'git commit -m "commit"'.format(dir=repos2_path))

    root.component.deploy()
    assert not os.path.exists(root.component.map('clone/bar'))
    git.url = repos2_path
    root.component.deploy()
    assert os.path.exists(root.component.map('clone/bar'))



from batou import output, DeploymentError
from collections import defaultdict
import contextlib
import fcntl
import hashlib
import itertools
import os
import pkg_resources
import socket
import subprocess
import sys
import time


def self_id():
    template = 'batou/{version} ({python}, {system})'
    system = os.uname()
    system = ' '.join([system[0], system[2], system[4]])
    version = pkg_resources.require("batou")[0].version
    python = sys.subversion[0]
    python += ' {0}.{1}.{2}-{3}{4}'.format(*sys.version_info)
    return template.format(**locals())


class MultiFile(object):

    def __init__(self, files):
        self.files = files

    def write(self, value):
        for file in self.files:
            file.write(value)

    def flush(self):
        for file in self.files:
            file.flush()


@contextlib.contextmanager
def locked(filename):
    # XXX can we make this not leave files around?
    with open(filename, 'a+') as lockfile:
        try:
            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            print >> sys.stderr, 'Could not acquire lock {}'.format(filename)
            raise RuntimeError(
                'cannot create lock "%s": more than one instance running '
                'concurrently?' % lockfile, lockfile)
        # publishing the process id comes handy for debugging
        lockfile.seek(0)
        lockfile.truncate()
        print >> lockfile, os.getpid()
        lockfile.flush()
        yield
        lockfile.seek(0)
        lockfile.truncate()


def flatten(list_of_lists):
    return list(itertools.chain.from_iterable(list_of_lists))


def notify_send(title, description):
    subprocess.call(['notify-send', title, description])


def notify_macosx(title, description):
    subprocess.call([
        'osascript', '-e',
        'display notification "{}" with title "{}"'.format(
            description, title)])


def notify_none(title, description):
    pass


try:
    subprocess.check_output(
        ['which', 'osascript'], stderr=subprocess.STDOUT)
    notify = notify_macosx
except (subprocess.CalledProcessError, OSError):
    try:
        subprocess.check_output(
            ['which', 'notify-send'], stderr=subprocess.STDOUT)
        notify = notify_send
    except (subprocess.CalledProcessError, OSError):
        notify = notify_none


resolve_override = {}
resolve_v6_override = {}


def resolve(host, resolve_override=resolve_override):
    address = resolve_override.get(host)
    if not address:
        try:
            address = socket.gethostbyname(host)
        except socket.gaierror:
            return None
    return address


def resolve_v6(host, port, resolve_override=resolve_v6_override):
    address = resolve_override.get(host)
    if not address:
        try:
            address = socket.getaddrinfo(
                host, int(port),
                socket.AF_INET6)[0][4][0]
        except socket.gaierror:
            address = None
    if address and address.startswith('fe80:'):
        # Don't hand out link-local addresses. This happes with
        # vagrant, and does not help as services cannot bind those
        # addresses without additional configuration, # i.e. the
        # interface.
        address = None
    return address


class Address(object):
    """An internet service address that can be listened and connected to.

    The constructor address is expected to be the address that can be
    connected to. The listen address will be computed automatically.

    .. code-block:: pycon

        >>> x = Address('localhost', 80)
        >>> str(x.connect)
        'localhost:80'
        >>> str(x.listen)
        '127.0.0.1:80'

    """

    #: The connect address as it should be used when configuring clients.
    #: This is a :py:class:`batou.utils.NetLoc` object.
    connect = None

    #: The listen (or bind) address as it should be used when configuring
    #: servers. This is a :py:class:`batou.utils.NetLoc` object.
    listen = None

    #: The IPv6 listen (or bind) address as it should be used when configuring
    #: servers. This is a :py:class:`batou.utils.NetLoc` object or None, if
    #: there is no IPv6 address.
    listen_v6 = None

    def __init__(self, connect_address, port=None):
        if ':' in connect_address:
            connect, port = connect_address.split(':')
        else:
            connect = connect_address
        if port is None:
            raise ValueError('Need port for service address.')
        self.connect = NetLoc(connect, str(port))
        v4_address = resolve(connect)
        if v4_address:
            self.listen = NetLoc(v4_address, str(port))
        v6_address = resolve_v6(connect, port)
        if v6_address:
            self.listen_v6 = NetLoc(v6_address, str(port))

        if not self.listen and not self.listen_v6:
            raise socket.gaierror("No v4 or v6 address for %s" % connect)

    def __lt__(self, other):
        if isinstance(other, Address):
            return str(self) < str(other)
        return NotImplemented

    def __str__(self):
        return str(self.connect)


class NetLoc(object):
    """A network location specified by host and port.

    Network locations can automatically render an appropriate string
    representation:

    .. code-block:: pycon

        >>> x = NetLoc('127.0.0.1')
        >>> x.host
        '127.0.0.1'
        >>> x.port
        None
        >>> str(x)
        '127.0.0.1'

        >>> y = NetLoc('127.0.0.1', 80)
        >>> str(y)
        '127.0.0.1:80'

    """

    #: The host part of this network location. Can be a hostname or IP address.
    host = None
    #: The port of this network location. Can be ``None`` or an integer.
    port = None

    def __init__(self, host, port=None):
        self.host = host
        self.port = port

    def __str__(self):
        if self.port:
            if ':' in self.host:  # ipv6
                fmt = '[{self.host}]:{self.port}'
            else:
                fmt = '{self.host}:{self.port}'
        else:
            fmt = '{self.host}'
        return fmt.format(self=self)


def revert_graph(graph):
    graph = ensure_graph_data(graph)
    reverse_graph = defaultdict(set)
    for node, dependencies in graph.items():
        # Ensure all nodes will exist
        reverse_graph[node]
        for dependency in dependencies:
            reverse_graph[dependency].add(node)
    return reverse_graph


def ensure_graph_data(graph):
    # Ensure that all nodes exist as keys even if they don't have outgoing
    # relations.
    for node, relations in list(graph.items()):
        for relation in relations:
            if relation not in graph:
                graph[relation] = set()
    return graph


class CycleError(ValueError):

    def __str__(self):
        message = []
        components = self.args[0].items()
        components.sort(key=lambda x: x[0].name)
        for component, subs in components:
            message.append(component.name + ' depends on')
            for sub in subs:
                message.append('        ' + sub.name)
        return '\n'.join(message)


def remove_nodes_without_outgoing_edges(graph):
    for node, dependencies in list(graph.items()):
        if not dependencies:
            del graph[node]


def topological_sort(graph):
    # Take a directed graph and provide a topological sort of all nodes.
    #
    # The graph is given as
    #
    # {node: [dependency, dependency], ...}
    #
    # If the graph has cycles a CycleError will be raised.

    graph = ensure_graph_data(graph)
    sorted = []
    reverse_graph = revert_graph(graph)
    roots = [node for node, incoming in reverse_graph.items()
             if not incoming]
    while roots:
        root = roots.pop()
        sorted.append(root)
        for node in list(graph[root]):
            graph[root].remove(node)
            reverse_graph[node].remove(root)
            if not reverse_graph[node]:
                roots.append(node)
    if any(graph.values()):
        # Simplify the graph a bit to make it easier to spot the cycle.
        remove_nodes_without_outgoing_edges(graph)
        raise CycleError(dict(graph))
    return sorted


class CmdExecutionError(DeploymentError, RuntimeError):

    def __init__(self, cmd, returncode, stdout, stderr):
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.args = (cmd, returncode, stdout, stderr)

    def report(self):
        output.error(self.cmd)
        output.tabular("Return code", str(self.returncode), red=True)
        output.line('STDOUT', red=True)
        output.annotate(self.stdout)
        output.line('STDERR', red=True)
        output.annotate(self.stderr)


def cmd(cmd, silent=False, ignore_returncode=False, communicate=True,
        env=None, acceptable_returncodes=[0]):
    if not isinstance(cmd, basestring):
        # We use `shell=True`, so the command needs to be a single string and
        # we need to pay attention to shell quoting.
        quoted_args = []
        for arg in cmd:
            arg = arg.replace('\'', '\\\'')
            if ' ' in arg:
                arg = "'{}'".format(arg)
            quoted_args.append(arg)
        cmd = ' '.join(quoted_args)
    if env is not None:
        add_to_env = env
        env = os.environ.copy()
        env.update(add_to_env)
    output.annotate('cmd: {}'.format(cmd), debug=True)
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True,
        env=env)
    if not communicate:
        # XXX See #12550
        return process
    stdout, stderr = process.communicate()
    if process.returncode not in acceptable_returncodes:
        if not ignore_returncode:
            raise CmdExecutionError(
                cmd, process.returncode, stdout, stderr)
    return stdout, stderr


class Timer(object):

    def __init__(self, note):
        self.duration = 0
        self.note = note

    def __enter__(self):
        self.started = time.time()

    def __exit__(self, exc1, exc2, exc3):
        self.duration = time.time() - self.started
        output.annotate(self.note + ' took %fs' % self.duration, debug=True)


def hash(path, function='md5'):
    h = getattr(hashlib, function)()
    for line in open(path):
        h.update(line)
    return h.hexdigest()



from zest.releaser.utils import execute_command
import logging
import os

logger = logging.getLogger(__name__)


def update_requirements(data):
    os.chdir(data['workingdir'])
    logger.info('Running buildout to update requirements.txt.')
    execute_command(['bin/buildout'])
    logger.info('Committing requirements.txt.')
    execute_command(['hg', 'commit', '-v', '-m', 'Update requirements.txt'])



from batou import UpdateNeeded
from batou.component import Component
from batou.lib.file import Directory
import os.path


class Checkout(Component):

    namevar = 'url'
    target = '.'
    revision = None

    def configure(self):
        self += Directory(self.target)

    def verify(self):
        with self.chdir(self.target):
            if not os.path.exists('.svn'):
                raise UpdateNeeded()
            stdout, stderr = self.cmd('svn info | grep Revision:')
            current_revision = stdout.replace('Revision:', '').strip()
            if current_revision != self.revision:
                raise UpdateNeeded()

    def update(self):
        with self.chdir(self.target):
            if not os.path.exists('.svn'):
                self.cmd(self.expand(
                    'svn co {{component.url}} . -r {{component.revision}}'))
            else:
                self.cmd(self.expand('svn revert -R .'))
                self.cmd(self.expand('svn switch {{component.url}}'))
                self.cmd(self.expand('svn up -r {{component.revision}}'))


Subversion = Checkout  # BBB



from batou import output
from batou.component import Component
import batou
import difflib
import glob
import os.path
import pwd
import shutil
import stat


def ensure_path_nonexistent(path):
    try:
        # cannot use os.path.exists(), since we also want to remove broken
        # symlinks
        os.lstat(path)
    except OSError:
        return
    if os.path.islink(path):
        os.unlink(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.unlink(path)


class File(Component):

    namevar = 'path'

    ensure = 'file'  # or: directory, symlink

    # Content oriented parameters
    content = None
    source = ''
    is_template = True
    template_context = None
    template_args = None  # dict, actually
    encoding = 'utf-8'

    # Unix attributes
    owner = None
    group = None
    mode = None

    # Symlink parameters
    link_to = ''

    # Leading directory creation
    leading = False

    def configure(self):
        self._unmapped_path = self.path
        self.path = self.map(self.path)
        if self.ensure == 'file':
            self += Presence(self.path, leading=self.leading)
        elif self.ensure == 'directory':
            self += Directory(self.path,
                              leading=self.leading,
                              source=self.source)
        elif self.ensure == 'symlink':
            self += Symlink(self.path, source=self.link_to)
        else:
            raise ValueError(
                'Ensure must be one of: file, directory, '
                'symlink not %s' % self.ensure)
        # variation: content or source explicitly given

        # The mode needs to be set early to allow batou to get out of
        # accidental "permission denied" situations.
        if self.mode:
            self += Mode(self.path, mode=self.mode)

        # no content or source given but file with same name
        # exists
        if self.ensure == 'file' and self.content is None and not self.source:
            guess_source = (
                self.root.defdir + '/' + os.path.basename(self.path))
            if os.path.isfile(guess_source):
                self.source = guess_source
            else:
                # Avoid the edge case where we want to support a very simple
                # case: specify File('asdf') and have an identical named file
                # in the component definition directory that will be templated
                # to the work directory.
                #
                # However, if you mis-spell the file, then you might
                # accidentally end up with an empty file in the work directory.
                # If you really want an empty File then you can either use
                # Presence(), or (recommended) use File('asdf', content='') to
                # make this explicit. We don't want to accidentally confuse the
                # convenience case (template a simple file) and an edge case
                # (have an empty file)
                raise ValueError(
                    "Missing implicit template file {}. Or did you want "
                    "to create an empty file? Then use File('{}', content='')."
                    .format(guess_source, self._unmapped_path))

        if self.content or self.source:
            if self.template_args is None:
                self.template_args = dict()
            if not self.template_context:
                self.template_context = self.parent
            content = Content(self.path,
                              source=self.source,
                              is_template=self.is_template,
                              template_context=self.template_context,
                              template_args=self.template_args,
                              encoding=self.encoding,
                              content=self.content)
            self += content
            self.content = content.content

        if self.owner:
            self += Owner(self.path, owner=self.owner)

        if self.group:
            self += Group(self.path, group=self.group)

    @property
    def namevar_for_breadcrumb(self):
        return os.path.relpath(self.path, self.environment.base_dir)

    def last_updated(self, key='st_mtime'):
        if not os.path.exists(self.path):
            return None
        return getattr(os.stat(self.path), key)


class BinaryFile(File):

    is_template = False
    encoding = None


class Presence(Component):

    namevar = 'path'
    leading = False

    def configure(self):
        self.path = self.map(self.path)
        if self.leading:
            self += Directory(os.path.dirname(self.path),
                              leading=self.leading)

    def verify(self):
        if not os.path.isfile(self.path):
            raise batou.UpdateNeeded()

    def update(self):
        ensure_path_nonexistent(self.path)
        with open(self.path, 'w'):
            # We're just touching it.
            pass

    @property
    def namevar_for_breadcrumb(self):
        return os.path.relpath(self.path, self.environment.base_dir)

    def last_updated(self, key='st_mtime'):
        if not os.path.exists(self.path):
            return None
        return getattr(os.stat(self.path), key)


class SyncDirectory(Component):

    namevar = 'path'
    source = None
    exclude = ()

    verify_opts = '-rclnv'
    sync_opts = '--inplace -lr'

    def configure(self):
        self.path = self.map(self.path)
        self.source = os.path.normpath(
            os.path.join(self.root.defdir, self.source))

    @property
    def exclude_arg(self):
        if not self.exclude:
            return ''
        return ' '.join("--exclude '{}'".format(x) for x in self.exclude) + ' '

    def verify(self):
        stdout, stderr = self.cmd('rsync {} {}{}/ {}'.format(
            self.verify_opts, self.exclude_arg, self.source, self.path))

        # In case of we see non-convergent rsync runs
        output.annotate('rsync result:', debug=True)
        output.annotate(stdout, debug=True)

        if len(stdout.strip().splitlines()) - 4 > 0:
            raise batou.UpdateNeeded()

    def update(self):
        self.cmd('rsync {} {}{}/ {}'.format(
            self.sync_opts, self.exclude_arg, self.source, self.path))

    @property
    def namevar_for_breadcrumb(self):
        return os.path.relpath(self.path, self.environment.base_dir)


class Directory(Component):

    namevar = 'path'
    leading = False
    source = None
    exclude = ()

    def configure(self):
        self.path = self.map(self.path)
        if self.source:
            # XXX The ordering is wrong. SyncDirectory should run *after*.
            self += SyncDirectory(
                self.path, source=self.source, exclude=self.exclude)

    def verify(self):
        if not os.path.isdir(self.path):
            raise batou.UpdateNeeded()

    def update(self):
        ensure_path_nonexistent(self.path)
        if self.leading:
            os.makedirs(self.path)
        else:
            os.mkdir(self.path)

    def last_updated(self, key='st_mtime'):
        newest = None
        for dirpath, dirnames, filenames in os.walk(self.path):
            for filename in filenames:
                time = getattr(os.stat(os.path.join(dirpath, filename)), key)
                if time > newest:
                    newest = time
        return newest

    @property
    def namevar_for_breadcrumb(self):
        return os.path.relpath(self.path, self.environment.base_dir)


class FileComponent(Component):

    namevar = 'path'
    leading = False

    def configure(self):
        self.original_path = self.path
        self.path = self.map(self.original_path)

    @property
    def namevar_for_breadcrumb(self):
        return os.path.relpath(self.path, self.environment.base_dir)


class Content(FileComponent):

    content = None
    is_template = File.is_template
    source = ''
    template_context = None
    template_args = None  # dict, actually

    # If content is given as unicode (always the case with templates)
    # then require it to be encodable. We assume UTF-8 as a sensible default
    # for most use cases and allow overrides.
    encoding = 'utf-8'

    _delayed = False

    def configure(self):
        super(Content, self).configure()

        # Step 1: Determine content attribute:
        # - it might be given directly (content='...'),
        # - we might have been passed a filename (source='...'), or
        # - we might fall back using the path attribute (namevar)
        if self.source and self.content:
            raise ValueError(
                'Only one of either "content" or "source" are allowed.')

        if not self.content:
            if not self.source:
                self.source = self.original_path

            if not self.source.startswith('/'):
                self.source = os.path.join(self.root.defdir, self.source)

        self._render()

    def _render(self):
        # Phase 1: acquire the source data into self.content
        if self.source:
            if os.path.exists(self.source):
                with open(self.source, 'r') as f:
                    self.content = f.read()
            else:
                if self._delayed:
                    raise RuntimeError(
                        'Could not find file {}'.format(self.source))
                # We need to try rendering again later.
                self._delayed = True
                return

        # Phase 2: Decode, if we have an encoding.
        if self.encoding and not isinstance(self.content, unicode):
            self.content = self.content.decode(self.encoding)

        # Phase 3: If we have a template, render it.
        if self.is_template:
            if self.template_args is None:
                self.template_args = dict()
            if not self.template_context:
                self.template_context = self.parent
            self.content = self.expand(
                self.content, self.template_context, args=self.template_args)

        # Phase 4: If we have an encoding, encode the content again
        if self.encoding:
            self.content = self.content.encode(self.encoding)

    def verify(self):
        if self._delayed:
            self._render()
        with open(self.path, 'r') as target:
            current = target.read()
            if current != self.content:
                for line in difflib.unified_diff(current.splitlines(),
                                                 self.content.splitlines()):
                    output.annotate(line, debug=True)
                raise batou.UpdateNeeded()

    def update(self):
        with open(self.path, 'w') as target:
            target.write(self.content)


class Owner(FileComponent):

    def verify(self):
        if isinstance(self.owner, str):
            self.owner = pwd.getpwnam(self.owner).pw_uid
        current = os.stat(self.path).st_uid
        if current != self.owner:
            raise batou.UpdateNeeded()

    def update(self):
        group = os.stat(self.path).st_gid
        os.chown(self.path, self.owner, group)


class Group(FileComponent):

    def configure(self):
        super(Group, self).configure()
        if isinstance(self.group, str):
            self.group = pwd.getpwnam(self.group)

    def verify(self):
        current = os.stat(self.path).st_gid
        if current != self.group:
            raise batou.UpdateNeeded()

    def update(self):
        owner = os.stat(self.path).st_uid
        os.chown(self.path, owner, self.group)


class Mode(FileComponent):

    def verify(self):
        try:
            self._select_stat_implementation()
        except AttributeError:
            # Happens on systems without lstat/lchmod implementation (like
            # Linux) Not sure whether ignoring it is really the right thing.
            return
        current = self._stat(self.path).st_mode
        if stat.S_IMODE(current) != self.mode:
            raise batou.UpdateNeeded()

    def update(self):
        self._chmod(self.path, self.mode)

    def _select_stat_implementation(self):
        self._stat = os.stat
        self._chmod = os.chmod
        if os.path.islink(self.path):
            self._stat = os.lstat
            self._chmod = os.lchmod


class Symlink(Component):

    namevar = 'target'

    def configure(self):
        self.target = self.map(self.target)
        self.source = self.map(self.source)

    def verify(self):
        if not os.path.islink(self.target):
            raise batou.UpdateNeeded()
        if os.readlink(self.target) != self.source:
            raise batou.UpdateNeeded()

    def update(self):
        ensure_path_nonexistent(self.target)
        os.symlink(self.source, self.target)


class Purge(Component):
    """Ensure that a set of files (given as a glob) does not exist."""

    namevar = 'pattern'

    def configure(self):
        self.pattern = self.map(self.pattern)

    def verify(self):
        if glob.glob(self.pattern):
            raise batou.UpdateNeeded()

    def update(self):
        for filename in glob.glob(self.pattern):
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            else:
                os.remove(filename)



# This code must not cause non-stdlib imports to support self-bootstrapping.
from ._output import output
import os.path
import traceback

__version__ = open(os.path.dirname(__file__) + '/version.txt').read().strip()


class UpdateNeeded(Exception):
    """A component requires an update."""


class ConfigurationError(Exception):
    """Indicates that an environment could not be configured successfully."""

    @property
    def sort_key(self):
        return (0, self.message)

    def __init__(self, message, component=None):
        self.message = message
        self.component = component

    def report(self):
        if self.component:
            message = '{}@{}: {}'.format(
                self.component.root.name,
                self.component.root.host.name,
                self.message)
        else:
            message = self.message
        output.error(message)


class ConversionError(ConfigurationError):
    """An override attribute could not be converted properly."""

    @property
    def sort_key(self):
        return (1, self.component.root.host.name,
                self.component._breadcrumbs, self.key)

    def __init__(self, component, key, value, conversion, error):
        self.component = component
        self.key = key
        self.value = value
        self.conversion = conversion
        self.error = error

    def report(self):
        output.error('Failed override attribute conversion')
        output.tabular(
            'Host', self.component.root.host.name,
            red=True)
        output.tabular(
            'Attribute',
            '{}.{}'.format(
                self.component._breadcrumbs,
                self.key),
            red=True)
        output.tabular('Conversion',
                       '{}({})'.format(self.conversion.__name__,
                                       repr(self.value)),
                       red=True)
        # TODO provide traceback in debug output
        output.tabular('Error', str(self.error), red=True)


class SilentConfigurationError(Exception):
    """These are exceptions that will be reported by other exceptions.

    They basically only influence control flow during configuration and
    are manually placed to avoid double reporting.

    """


class MissingOverrideAttributes(ConfigurationError):

    @property
    def sort_key(self):
        return (1, self.component.root.host.name,
                self.component._breadcrumbs, 0)

    def __init__(self, component, attributes):
        self.component = component
        self.attributes = attributes

    def report(self):
        output.error('Overrides for undefined attributes')
        output.tabular(
            'Host', self.component.root.host.name,
            red=True)
        output.tabular(
            'Component',
            self.component._breadcrumbs,
            red=True)
        output.tabular(
            'Attributes',
            ', '.join(self.attributes),
            red=True)

        # TODO point to specific line in secrets or environments
        # cfg file and show context


class DuplicateComponent(ConfigurationError):

    @property
    def sort_key(self):
        return (1, self.component.a.host.name, self.a.__name__)

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def report(self):
        output.error('Duplicate component "{}"'.format(self.a.__name_))
        output.tabular('Occurence', self.a.__file__)
        output.tabular('Occurence', self.b.__file__)


class UnknownComponentConfigurationError(ConfigurationError):
    """An unknown error occured while configuring a component."""

    @property
    def sort_key(self):
        return (1, self.root.host.name, 2)

    def __init__(self, root, exception, tb):
        self.root = root
        self.exception = exception
        stack = traceback.extract_tb(tb)
        from batou import environment, component
        while True:
            # Delete remoting-internal stack frames.'
            line = stack.pop(0)
            if line[0] in ['<string>', '<remote exec>',
                           environment.__file__.rstrip('c'),
                           component.__file__.rstrip('c')]:
                continue
            stack.insert(0, line)
            break
        self.traceback = ''.join(traceback.format_list(stack))

    def report(self):
        output.error(repr(self.exception))
        output.annotate(
            "This /might/ be a batou bug. Please consider reporting it.\n",
            red=True)
        output.tabular(
            "Host", self.root.host.name, red=True)
        output.tabular(
            "Component", self.root.name + '\n', red=True)
        output.annotate('Traceback (simplified, most recent call last):',
                        red=True)
        output.annotate(self.traceback, red=True)


class UnusedResources(ConfigurationError):
    """Some provided resources were never used."""

    @property
    def sort_key(self):
        return (3, 'unused')

    def __init__(self, resources):
        self.resources = resources

    def report(self):
        output.error("Unused provided resources")
        for key in sorted(self.resources):
            for component, value in self.resources[key].items():
                output.line(
                    '    Resource "{}" provided by {} with value {}'.format(
                        key, component.name, value),
                    red=True)


class UnsatisfiedResources(ConfigurationError):
    """Some required resources were never provided."""

    @property
    def sort_key(self):
        return (4, 'unsatisfied')

    def __init__(self, resources):
        self.resources = resources

    def report(self):
        output.error("Unsatisfied resource requirements")
        for key in sorted(self.resources):
            output.line('    Resource "{}" required by {}'.format(
                key, ','.join(r.name for r in self.resources[key])),
                red=True)


class MissingEnvironment(ConfigurationError):
    """The specified environment does not exist."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, environment):
        self.environment = environment

    def report(self):
        output.error("Missing environment")
        output.tabular('Environment', self.environment.name, red=True)


class ComponentLoadingError(ConfigurationError):
    """The specified component file failed to load."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, filename, exception):
        self.filename = filename
        self.exception = exception

    def report(self):
        output.error("Failed loading component file")
        output.tabular("File", self.filename, red=True)
        output.tabular("Exception", str(self.exception), red=True)
        # TODO provide traceback in debug output


class MissingComponent(ConfigurationError):
    """The specified environment does not exist."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, component, hostname):
        self.component = component
        self.hostname = hostname

    def report(self):
        output.error("Missing component")
        output.tabular('Component', self.component, red=True)
        output.tabular('Host', self.hostname, red=True)


class SuperfluousSection(ConfigurationError):
    """A superfluous section was found in the environment
    configuration file."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, section):
        self.section = section

    def report(self):
        output.error("Superfluous section in environment configuration")
        output.tabular("Section", self.section, red=True)
        # TODO provide location and context in debug output


class SuperfluousComponentSection(ConfigurationError):
    """A component section was found in the environment
    but no associated component is known."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, component):
        self.component = component

    def report(self):
        output.error("Override section for unknown component found")
        output.tabular("Component", self.component, red=True)
        # TODO provide traceback in debug output


class SuperfluousSecretsSection(ConfigurationError):
    """A component section was found in the secrets
    but no associated component is known."""

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, component):
        self.component = component

    def report(self):
        output.error("Secrets section for unknown component found")
        output.tabular("Component", self.component, red=True)
        # TODO provide traceback in debug output


class CycleErrorDetected(ConfigurationError):
    """We think we found a cycle in the component dependencies.
    """

    @property
    def sort_key(self):
        return (99, )

    def __init__(self, error):
        self.error = error

    def report(self):
        output.error("Found dependency cycle")
        output.annotate(str(self.error), red=True)
        # TODO provide traceback in debug output


class NonConvergingWorkingSet(ConfigurationError):
    """A working set did not converge."""

    @property
    def sort_key(self):
        return (100, )

    def __init__(self, roots):
        self.roots = roots

    def report(self):
        # TODO show this last or first, but not in the middle
        # of everything
        output.error("{} remaining unconfigured component(s)".format(
                     len(self.roots)))
        # TODO show all incl. their host name in -vv or so
        # output.annotate(', '.join(c.name for c in self.roots))


class DeploymentError(Exception):
    """Indicates that a deployment failed.."""

    @property
    def sort_key(self):
        return (200, )

    def report(self):
        pass


class RepositoryDifferentError(DeploymentError):
    """The repository on the remote side is different."""

    @property
    def sort_key(self):
        return (150, )

    def __init__(self, local, remote):
        self.local = local
        self.remote = remote

    def report(self):
        output.error(
            "The remote working copy is based on a different revision. "
            "Maybe you tried to deploy from the wrong branch.")
        output.tabular("Local", self.local, red=True)
        output.tabular("Remote", self.remote, red=True)


class DuplicateHostError(ConfigurationError):

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, hostname):
        self.hostname = hostname

    def report(self):
        output.error("Duplicate definition of host: {}".format(self.hostname))


class InvalidIPAddressError(ConfigurationError):

    @property
    def sort_key(self):
        return (0, )

    def __init__(self, address):
        self.address = address

    def report(self):
        output.error("Not a valid IP address: {}".format(self.address))



from batou.secrets.encryption import EncryptedConfigFile as BaseEncConfigFile
from batou.secrets.encryption import NEW_FILE_TEMPLATE
import ConfigParser
import os
import os.path
import pytest
import shutil
import subprocess
import tempfile


FIXTURE = os.path.join(os.path.dirname(__file__), 'fixture')
cleartext_file = os.path.join(FIXTURE, 'cleartext.cfg')
encrypted_file = os.path.join(FIXTURE, 'encrypted.cfg')


@pytest.fixture(scope='session', autouse=True)
def cleanup_gpg_sockets():
    yield
    for path in [
            'S.dirmngr',
            'S.gpg-agent',
            'S.gpg-agent.browser',
            'S.gpg-agent.extra',
            'S.gpg-agent.ssh']:
        try:
            os.remove(os.path.join(FIXTURE, 'gnupg', path))
        except OSError:
            pass


class EncryptedConfigFile(BaseEncConfigFile):

    gpg_homedir = os.path.join(FIXTURE, 'gnupg')
    gpg_opts = (BaseEncConfigFile.gpg_opts +
                ' --homedir {}'.format(gpg_homedir))


def test_decrypt():
    with EncryptedConfigFile(encrypted_file) as secrets:
        with open(cleartext_file) as cleartext:
            assert cleartext.read().strip() == secrets.read().strip()


def test_caches_cleartext():
    with EncryptedConfigFile(encrypted_file) as secrets:
        secrets.read()
        secrets._cleartext = '[foo]bar=2'
        assert secrets.read() == '[foo]bar=2'


def test_decrypt_missing_key():
    secrets = EncryptedConfigFile(
        encrypted_file)
    secrets.gpg_opts = BaseEncConfigFile.gpg_opts + ' --homedir /tmp'
    with pytest.raises(subprocess.CalledProcessError):
        f = secrets.__enter__()
        f.read()


def test_write_should_fail_unless_write_locked():
    with EncryptedConfigFile(encrypted_file) as secrets:
        with pytest.raises(RuntimeError):
            secrets.write('dummy')


def test_open_nonexistent_file_for_read_should_fail():
    with pytest.raises(IOError):
        EncryptedConfigFile('/no/such/file').__enter__()


def test_open_nonexistent_file_for_write_should_create_template_file():
    tf = tempfile.NamedTemporaryFile(prefix='new_encrypted.')
    tf.close()  # deletes file
    encrypted = EncryptedConfigFile(
        tf.name, write_lock=True)
    with encrypted as secrets:
        assert secrets.read() == NEW_FILE_TEMPLATE
        assert os.path.exists(secrets.encrypted_file)
    os.unlink(tf.name)


def test_write_unparsable_raises_error():
    with tempfile.NamedTemporaryFile(prefix='new_encrypted.') as tf:
        shutil.copy(encrypted_file, tf.name)
        encrypted = EncryptedConfigFile(
            tf.name, write_lock=True)
        with encrypted as secrets:
            with pytest.raises(ConfigParser.Error):
                secrets.write('some new file contents\n')


def test_write():
    with tempfile.NamedTemporaryFile(prefix='new_encrypted.') as tf:
        shutil.copy(encrypted_file, tf.name)
        encrypted = EncryptedConfigFile(
            tf.name, write_lock=True)
        with encrypted as secrets:
            secrets.write("""\
[batou]
members = batou
[asdf]
x = 1
""")

        with open(encrypted_file, 'rb') as old:
            with open(tf.name, 'rb') as new:
                assert old.read() != new.read()
        assert 0 != os.stat(tf.name).st_size


def test_write_fails_without_recipients():
    with tempfile.NamedTemporaryFile(prefix='new_encrypted.') as tf:
        shutil.copy(encrypted_file, tf.name)
        encrypted = EncryptedConfigFile(
            tf.name, write_lock=True)
        with encrypted as secrets:
            with pytest.raises(ValueError):
                secrets.write("""\
[batou]
members =
[asdf]
x = 1
""")


def test_write_fails_if_recipient_key_is_missing_keeps_old_file():
    with tempfile.NamedTemporaryFile(prefix='new_encrypted.') as tf:
        shutil.copy(encrypted_file, tf.name)
        encrypted = EncryptedConfigFile(
            tf.name, write_lock=True)
        with encrypted as secrets:
            with pytest.raises(RuntimeError):
                secrets.write("""\
[batou]
members = foobar@example.com
[asdf]
x = 1
""")
        assert open(tf.name).read() == open(encrypted_file).read()



from batou.component import Component
from batou.lib.file import File


class ZEO(Component):

    def configure(self):
        self.provide('zeo', '127.0.0.1:9000')
        # As reported in #11183 we had a regression that caused the
        # sub-component to remove the provided values accidentally.
        self += File('asdf', content='asdf')
